@@ -1,3 +1,6 @@
+// Copyright 2024 Bjørn Erik Pedersen
+// SPDX-License-Identifier: MIT
+
 package imagemeta_test
 
 import (
@@ -5,6 +8,7 @@ import (
 	"errors"
 	"fmt"
 	"io"
+	"maps"
 	"math"
 	"math/rand"
 	"os"
@@ -14,6 +18,7 @@ import (
 	"strconv"
 	"strings"
 	"testing"
+	"time"
 
 	"github.com/bep/imagemeta"
 	"github.com/rwcarlsen/goexif/exif"
@@ -57,7 +62,8 @@ func TestDecodeAllImageFormats(t *testing.T) {
 
 func TestDecodeWebP(t *testing.T) {
 	c := qt.New(t)
-	tags := extractTags(t, "sunrise.webp", imagemeta.EXIF|imagemeta.IPTC|imagemeta.XMP)
+	tags, err := extractTags(t, "sunrise.webp", imagemeta.EXIF|imagemeta.IPTC|imagemeta.XMP)
+	c.Assert(err, qt.IsNil)
 
 	c.Assert(tags.EXIF()["Copyright"].Value, qt.Equals, "Bjørn Erik Pedersen")
 	c.Assert(tags.EXIF()["ApertureValue"].Value, eq, 5.6)
@@ -66,6 +72,12 @@ func TestDecodeWebP(t *testing.T) {
 	c.Assert(tags.XMP()["City"].Value, qt.Equals, "Benalmádena")
 }
 
+// For development.
+func TestDecodeAdhoc(t *testing.T) {
+	t.Skip("used in development")
+	extractTags(t, "/Users/bep/dump/summer.webp", imagemeta.EXIF)
+}
+
 func TestDecodeJPEG(t *testing.T) {
 	c := qt.New(t)
 
@@ -73,7 +85,8 @@ func TestDecodeJPEG(t *testing.T) {
 		return true
 	}
 
-	tags := extractTagsWithFilter(t, "sunrise.jpg", imagemeta.EXIF|imagemeta.IPTC|imagemeta.XMP, shouldInclude)
+	tags, err := extractTagsWithFilter(t, "sunrise.jpg", imagemeta.EXIF|imagemeta.IPTC|imagemeta.XMP, shouldInclude)
+	c.Assert(err, qt.IsNil)
 
 	c.Assert(tags.EXIF()["Copyright"].Value, qt.Equals, "Bjørn Erik Pedersen")
 	c.Assert(tags.EXIF()["ApertureValue"].Value, eq, 5.6)
@@ -89,7 +102,8 @@ func TestDecodePNG(t *testing.T) {
 		return true
 	}
 
-	tags := extractTagsWithFilter(t, "sunrise.png", imagemeta.EXIF|imagemeta.IPTC|imagemeta.XMP, shouldInclude)
+	tags, err := extractTagsWithFilter(t, "sunrise.png", imagemeta.EXIF|imagemeta.IPTC|imagemeta.XMP, shouldInclude)
+	c.Assert(err, qt.IsNil)
 
 	c.Assert(len(tags.EXIF()), qt.Equals, 61)
 	c.Assert(len(tags.IPTC()), qt.Equals, 14)
@@ -109,7 +123,8 @@ func TestThumbnailOffset(t *testing.T) {
 	}
 
 	offset := func(filename string) uint32 {
-		tags := extractTagsWithFilter(t, filename, imagemeta.EXIF, shouldHandle)
+		tags, err := extractTagsWithFilter(t, filename, imagemeta.EXIF, shouldHandle)
+		c.Assert(err, qt.IsNil)
 		return tags.EXIF()["ThumbnailOffset"].Value.(uint32)
 	}
 
@@ -122,7 +137,8 @@ func TestThumbnailOffset(t *testing.T) {
 func TestDecodeTIFF(t *testing.T) {
 	c := qt.New(t)
 
-	tags := extractTags(t, "sunrise.tif", imagemeta.EXIF|imagemeta.IPTC|imagemeta.XMP)
+	tags, err := extractTags(t, "sunrise.tif", imagemeta.EXIF|imagemeta.IPTC|imagemeta.XMP)
+	c.Assert(err, qt.IsNil)
 
 	c.Assert(len(tags.EXIF()), qt.Equals, 76)
 	c.Assert(len(tags.XMP()), qt.Equals, 146)
@@ -136,8 +152,9 @@ func TestDecodeTIFF(t *testing.T) {
 func TestDecodeCorrupt(t *testing.T) {
 	c := qt.New(t)
 
-	files, err := filepath.Glob(filepath.Join("testdata", "corrupt", "*.*"))
+	files, err := filepath.Glob(filepath.Join("testdata", "images", "corrupt", "*.*"))
 	c.Assert(err, qt.IsNil)
+	c.Assert(files, qt.HasLen, 3)
 
 	for _, file := range files {
 		img, err := os.Open(file)
@@ -147,11 +164,46 @@ func TestDecodeCorrupt(t *testing.T) {
 			return nil
 		}
 		err = imagemeta.Decode(imagemeta.Options{R: img, ImageFormat: format, HandleTag: handleTag, Warnf: panicWarnf})
-		c.Assert(imagemeta.IsInvalidFormat(err), qt.IsTrue, qt.Commentf("file: %s", file))
+
+		if !imagemeta.IsInvalidFormat(err) {
+			c.Assert(err, qt.ErrorMatches, "UserComment: expected \\[\\]uint8, got string", qt.Commentf("file: %s", file))
+		}
+
 		img.Close()
 	}
 }
 
+func TestDecodeUserCommentWithInvalidEncoding(t *testing.T) {
+	c := qt.New(t)
+
+	img, err := os.Open(filepath.Join("testdata", "images", "invalid-encoding-usercomment.jpg"))
+	c.Assert(err, qt.IsNil)
+	defer img.Close()
+
+	var userComment any
+	handleTag := func(ti imagemeta.TagInfo) error {
+		if ti.Tag == "UserComment" {
+			userComment = ti.Value
+		}
+		return nil
+	}
+
+	var warning error
+	warnf := func(format string, args ...any) {
+		warning = fmt.Errorf(format, args...)
+	}
+
+	err = imagemeta.Decode(imagemeta.Options{R: img, ImageFormat: imagemeta.JPEG, HandleTag: handleTag, Sources: imagemeta.EXIF, Warnf: warnf})
+	c.Assert(err, qt.IsNil)
+
+	// Expect user comment to be decoded
+	c.Assert(userComment, qt.IsNotNil)
+	c.Assert(userComment, eq, "UserComment")
+
+	// But with a warning about incorrect type
+	c.Assert(warning, qt.ErrorMatches, "UserComment: expected \\[\\]uint8, got string")
+}
+
 func TestDecodeCustomXMPHandler(t *testing.T) {
 	c := qt.New(t)
 
@@ -204,7 +256,7 @@ func TestDecodeShouldHandleTagEXIF(t *testing.T) {
 
 	const numTagsTotal = 64
 
-	for i := 0; i < 30; i++ {
+	for range 30 {
 		img, close := getSunrise(c, imagemeta.JPEG)
 		c.Cleanup(close)
 
@@ -297,7 +349,8 @@ func TestDecodeNamespace(t *testing.T) {
 		return true
 	}
 
-	tags := extractTagsWithFilter(t, "sunrise.jpg", imagemeta.EXIF|imagemeta.IPTC|imagemeta.XMP, shouldInclude)
+	tags, err := extractTagsWithFilter(t, "sunrise.jpg", imagemeta.EXIF|imagemeta.IPTC|imagemeta.XMP, shouldInclude)
+	c.Assert(err, qt.IsNil)
 
 	c.Assert(tags.EXIF()["Artist"].Namespace, qt.Equals, "IFD0")
 	c.Assert(tags.EXIF()["GPSLatitude"].Namespace, qt.Equals, "IFD0/GPSInfoIFD")
@@ -366,10 +419,25 @@ func TestDecodeIPTCOrientationOnly(t *testing.T) {
 	c.Assert(len(tags.IPTC()), qt.Equals, 1)
 }
 
+func TestDecodeLargeExifTimeout(t *testing.T) {
+	c := qt.New(t)
+
+	withOpts := func(opts *imagemeta.Options) {
+		opts.Timeout = time.Duration(500 * time.Millisecond)
+
+		// Set the limits to something high to make sure we time out.
+		opts.LimitNumTags = 1000000
+		opts.LimitTagSize = 10000000
+	}
+	_, err := extractTags(t, "largeexif.png", imagemeta.EXIF, withOpts)
+	c.Assert(err, qt.ErrorMatches, "timed out after 500ms")
+}
+
 func TestDecodeXMPJPG(t *testing.T) {
 	c := qt.New(t)
 
-	tags := extractTags(t, "sunrise.jpg", imagemeta.XMP)
+	tags, err := extractTags(t, "sunrise.jpg", imagemeta.XMP)
+	c.Assert(err, qt.IsNil)
 
 	c.Assert(len(tags.EXIF()) == 0, qt.IsTrue)
 	c.Assert(len(tags.IPTC()) == 0, qt.IsTrue)
@@ -388,6 +456,10 @@ func TestGoldenEXIFHugoIssue12669(t *testing.T) {
 	compareWithExiftoolOutput(t, "hugo-issue-12669.jpg", imagemeta.EXIF)
 }
 
+func TestGoldenEXIFIssue34(t *testing.T) {
+	compareWithExiftoolOutput(t, "outofbounds-issue-34.jpg", imagemeta.EXIF)
+}
+
 func TestGoldenEXIF(t *testing.T) {
 	withGolden(t, imagemeta.EXIF)
 }
@@ -423,14 +495,16 @@ func TestGoldenTagCountXMP(t *testing.T) {
 func TestLatLong(t *testing.T) {
 	c := qt.New(t)
 
-	tags := extractTags(t, "sunrise.jpg", imagemeta.EXIF)
+	tags, err := extractTags(t, "sunrise.jpg", imagemeta.EXIF)
+	c.Assert(err, qt.IsNil)
 
 	lat, long, err := tags.GetLatLong()
 	c.Assert(err, qt.IsNil)
 	c.Assert(lat, eq, float64(36.59744166))
 	c.Assert(long, eq, float64(-4.50846))
 
-	tags = extractTags(t, "goexif/geodegrees_as_string.jpg", imagemeta.EXIF)
+	tags, err = extractTags(t, "goexif/geodegrees_as_string.jpg", imagemeta.EXIF)
+	c.Assert(err, qt.IsNil)
 	lat, long, err = tags.GetLatLong()
 	c.Assert(err, qt.IsNil)
 	c.Assert(lat, eq, float64(52.013888888))
@@ -440,7 +514,8 @@ func TestLatLong(t *testing.T) {
 func TestGetDateTime(t *testing.T) {
 	c := qt.New(t)
 
-	tags := extractTags(t, "sunrise.jpg", imagemeta.EXIF)
+	tags, err := extractTags(t, "sunrise.jpg", imagemeta.EXIF)
+	c.Assert(err, qt.IsNil)
 	d, err := tags.GetDateTime()
 	c.Assert(err, qt.IsNil)
 	c.Assert(d.Format("2006-01-02"), qt.Equals, "2017-10-27")
@@ -498,7 +573,8 @@ func assertGoldenInfoTagCount(t testing.TB, filename string, sources imagemeta.S
 		return true
 	}
 
-	tags := extractTagsWithFilter(t, filename, sources, shouldHandle)
+	tags, err := extractTagsWithFilter(t, filename, sources, shouldHandle)
+	c.Assert(err, qt.IsNil)
 	all := tags.All()
 
 	// Our XMP parsing is currently a little limited so be a little lenient with the assertions.
@@ -511,28 +587,16 @@ func assertGoldenInfoTagCount(t testing.TB, filename string, sources imagemeta.S
 	tagsRight := make(map[string]any)
 
 	if sources.Has(imagemeta.EXIF) {
-		for k, v := range tags.EXIF() {
-			tagsLeft[k] = v
-		}
-		for k, v := range goldenInfo.EXIF {
-			tagsRight[k] = v
-		}
+		maps.Copy(tagsLeft, tags.EXIF())
+		maps.Copy(tagsRight, goldenInfo.EXIF)
 	}
 	if sources.Has(imagemeta.IPTC) {
-		for k, v := range tags.IPTC() {
-			tagsLeft[k] = v
-		}
-		for k, v := range goldenInfo.IPTC {
-			tagsRight[k] = v
-		}
+		maps.Copy(tagsLeft, tags.IPTC())
+		maps.Copy(tagsRight, goldenInfo.IPTC)
 	}
 	if sources.Has(imagemeta.XMP) {
-		for k, v := range tags.XMP() {
-			tagsLeft[k] = v
-		}
-		for k, v := range goldenInfo.XMP {
-			tagsRight[k] = v
-		}
+		maps.Copy(tagsLeft, tags.XMP())
+		maps.Copy(tagsRight, goldenInfo.XMP)
 	}
 
 	count := 0
@@ -585,7 +649,8 @@ func assertGoldenInfoTagCount(t testing.TB, filename string, sources imagemeta.S
 
 func compareWithExiftoolOutput(t testing.TB, filename string, sources imagemeta.Source) {
 	c := qt.New(t)
-	tags := extractTags(t, filename, sources)
+	tags, err := extractTags(t, filename, sources)
+	c.Assert(err, qt.IsNil)
 	all := tags.All()
 	tagsGolden := readGoldenInfo(t, filename)
 
@@ -612,9 +677,6 @@ func compareWithExiftoolOutput(t testing.TB, filename string, sources imagemeta.
 			case imagemeta.Rat[int32]:
 				return v.Float64()
 			case float64:
-				if math.IsInf(v, 1) {
-					return "undef"
-				}
 				return v
 			case int64:
 				return float64(v)
@@ -701,6 +763,17 @@ func compareWithExiftoolOutput(t testing.TB, filename string, sources imagemeta.
 		if found {
 			expect := normalizeThem(v.Tag, exifToolValue, v.Source)
 			got := normalizeUs(v.Tag, v.Value, v.Source)
+			if v, ok := got.(float64); ok {
+				if math.IsInf(v, 1) {
+					panic(fmt.Errorf("inf: %v", v))
+				}
+				if math.IsInf(v, -1) {
+					panic(fmt.Errorf("-inf: %v", v))
+				}
+				if math.IsNaN(v) {
+					panic(fmt.Errorf("nan: %v", v))
+				}
+			}
 			c.Assert(got, eq, expect, qt.Commentf("%s (%s): got: %T/%T %v %q\n\n%s\n\n%s", v.Tag, v.Source, got, expect, v.Value, filename, got, expect))
 		}
 	}
@@ -721,17 +794,22 @@ func extToFormat(ext string) imagemeta.ImageFormat {
 	}
 }
 
-func extractTags(t testing.TB, filename string, sources imagemeta.Source) imagemeta.Tags {
+func extractTags(t testing.TB, filename string, sources imagemeta.Source, opts ...withOptions) (imagemeta.Tags, error) {
 	shouldHandle := func(ti imagemeta.TagInfo) bool {
 		// Drop the thumbnail tags.
 		return ti.Namespace != "IFD1"
 	}
-	return extractTagsWithFilter(t, filename, sources, shouldHandle)
+	return extractTagsWithFilter(t, filename, sources, shouldHandle, opts...)
 }
 
-func extractTagsWithFilter(t testing.TB, filename string, sources imagemeta.Source, shouldHandle func(ti imagemeta.TagInfo) bool) imagemeta.Tags {
+type withOptions func(opts *imagemeta.Options)
+
+func extractTagsWithFilter(t testing.TB, filename string, sources imagemeta.Source, shouldHandle func(ti imagemeta.TagInfo) bool, opts ...withOptions) (imagemeta.Tags, error) {
 	t.Helper()
-	f, err := os.Open(filepath.Join("testdata", "images", filename))
+	if !filepath.IsAbs(filename) {
+		filename = filepath.Join("testdata", "images", filename)
+	}
+	f, err := os.Open(filename)
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -757,11 +835,30 @@ func extractTagsWithFilter(t testing.TB, filename string, sources imagemeta.Sour
 		panic(errors.New(s))
 	}
 
-	err = imagemeta.Decode(imagemeta.Options{R: f, ImageFormat: imageFormat, ShouldHandleTag: shouldHandle, HandleTag: handleTag, Warnf: warnf, Sources: sources})
-	if err != nil {
-		t.Fatal(fmt.Errorf("failed to decode %q: %w", filename, err))
+	imgOpts := imagemeta.Options{R: f, ImageFormat: imageFormat, ShouldHandleTag: shouldHandle, HandleTag: handleTag, Warnf: warnf, Sources: sources}
+	for _, opt := range opts {
+		opt(&imgOpts)
 	}
-	return tags
+
+	err = imagemeta.Decode(imgOpts)
+	if err != nil {
+		return tags, err
+	}
+
+	// See https://github.com/gohugoio/hugo/issues/12741 and https://github.com/golang/go/issues/59627
+	// Verify that it can be marshaled to JSON.
+	_, err = json.Marshal(tags.All())
+	if err != nil {
+		for tag, ti := range tags.All() {
+			_, err2 := json.Marshal(ti.Value)
+			if err2 != nil {
+				t.Fatal(fmt.Errorf("failed to marshal tag %q in source %s with value %v/%T to JSON: %w", tag, ti.Source, ti.Value, ti.Value, err2))
+			}
+		}
+		t.Fatal(fmt.Errorf("failed to marshal tags in %q to JSON: %w", filename, err))
+	}
+
+	return tags, nil
 }
 
 func readGoldenInfo(t testing.TB, filename string) goldenFileInfo {
@@ -822,7 +919,8 @@ func withTestDataFile(t testing.TB, fn func(path string, info os.FileInfo, err e
 }
 
 var goldenSkip = map[string]bool{
-	"goexif/geodegrees_as_string.jpg": true, // The file has many EXIF errors. I think we do a better job than exiftools, but there are some differences.
+	"goexif/geodegrees_as_string.jpg":  true, // The file has many EXIF errors. I think we do a better job than exiftools, but there are some differences.
+	"invalid-encoding-usercomment.jpg": true, // The file has an EXIF error that produces a warning in imagemeta. It's tested separately.
 }
 
 var isSpaceDelimitedFloatRe = regexp.MustCompile(`^(\d+\.\d+) (\d+\.\d+)`)
@@ -859,7 +957,7 @@ var eq = qt.CmpEquals(
 			if len(floatStringLeft) != len(floatStringRight) {
 				return false
 			}
-			for i := 0; i < len(floatStringLeft); i++ {
+			for i := range floatStringLeft {
 				left, err := strconv.ParseFloat(floatStringLeft[i], 64)
 				if err != nil {
 					return false
@@ -1044,6 +1142,6 @@ func BenchmarkDecodeCompareWithGoexif(b *testing.B) {
 	})
 }
 
-func panicWarnf(format string, args ...interface{}) {
+func panicWarnf(format string, args ...any) {
 	panic(fmt.Errorf(format, args...))
 }
