@@ -1,27 +1,28 @@
-package openapi3_test
+package openapi3
 
 import (
 	"context"
 	"encoding/base64"
 	"encoding/json"
 	"math"
+	"reflect"
 	"strings"
 	"testing"
 
-	"github.com/getkin/kin-openapi/openapi3"
 	"github.com/stretchr/testify/require"
+	"gopkg.in/yaml.v3"
 )
 
 type schemaExample struct {
 	Title         string
-	Schema        *openapi3.Schema
+	Schema        *Schema
 	Serialization interface{}
 	AllValid      []interface{}
 	AllInvalid    []interface{}
 }
 
 func TestSchemas(t *testing.T) {
-	openapi3.DefineStringFormat("uuid", openapi3.FormatOfStringForUUIDOfRFC4122)
+	DefineStringFormat("uuid", FormatOfStringForUUIDOfRFC4122)
 	for _, example := range schemaExamples {
 		t.Run(example.Title, testSchema(t, example))
 	}
@@ -36,43 +37,61 @@ func testSchema(t *testing.T, example schemaExample) func(*testing.T) {
 			jsonSchema, err := json.Marshal(schema)
 			require.NoError(t, err)
 			require.JSONEq(t, string(jsonSerialized), string(jsonSchema))
-			var dataUnserialized openapi3.Schema
+			var dataUnserialized Schema
 			err = json.Unmarshal(jsonSerialized, &dataUnserialized)
 			require.NoError(t, err)
-			var dataSchema openapi3.Schema
+			var dataSchema Schema
 			err = json.Unmarshal(jsonSchema, &dataSchema)
 			require.NoError(t, err)
 			require.Equal(t, dataUnserialized, dataSchema)
 		}
-		for _, value := range example.AllValid {
-			err := validateSchema(t, schema, value)
-			require.NoError(t, err)
-		}
-		for _, value := range example.AllInvalid {
-			err := validateSchema(t, schema, value)
-			require.Error(t, err)
+		for validateFuncIndex, validateFunc := range validateSchemaFuncs {
+			for index, value := range example.AllValid {
+				err := validateFunc(t, schema, value)
+				require.NoErrorf(t, err, "ValidateFunc #%d, AllValid #%d: %#v", validateFuncIndex, index, value)
+			}
+			for index, value := range example.AllInvalid {
+				err := validateFunc(t, schema, value)
+				require.Errorf(t, err, "ValidateFunc #%d, AllInvalid #%d: %#v", validateFuncIndex, index, value)
+			}
 		}
 		// NaN and Inf aren't valid JSON but are handled
-		for _, value := range []interface{}{math.NaN(), math.Inf(-1), math.Inf(+1)} {
+		for index, value := range []interface{}{math.NaN(), math.Inf(-1), math.Inf(+1)} {
 			err := schema.VisitJSON(value)
-			require.Error(t, err)
+			require.Errorf(t, err, "NaNAndInf #%d: %#v", index, value)
 		}
 	}
 }
 
-func validateSchema(t *testing.T, schema *openapi3.Schema, value interface{}) error {
+func validateSchemaJSON(t *testing.T, schema *Schema, value interface{}, opts ...SchemaValidationOption) error {
 	data, err := json.Marshal(value)
 	require.NoError(t, err)
 	var val interface{}
 	err = json.Unmarshal(data, &val)
 	require.NoError(t, err)
-	return schema.VisitJSON(val)
+	return schema.VisitJSON(val, opts...)
+}
+
+func validateSchemaYAML(t *testing.T, schema *Schema, value interface{}, opts ...SchemaValidationOption) error {
+	data, err := yaml.Marshal(value)
+	require.NoError(t, err)
+	var val interface{}
+	err = yaml.Unmarshal(data, &val)
+	require.NoError(t, err)
+	return schema.VisitJSON(val, opts...)
+}
+
+type validateSchemaFunc func(t *testing.T, schema *Schema, value interface{}, opts ...SchemaValidationOption) error
+
+var validateSchemaFuncs = []validateSchemaFunc{
+	validateSchemaJSON,
+	validateSchemaYAML,
 }
 
 var schemaExamples = []schemaExample{
 	{
 		Title:         "EMPTY SCHEMA",
-		Schema:        &openapi3.Schema{},
+		Schema:        &Schema{},
 		Serialization: map[string]interface{}{
 			// This OA3 schema is exactly this draft-04 schema:
 			//   {"not": {"type": "null"}}
@@ -92,7 +111,7 @@ var schemaExamples = []schemaExample{
 
 	{
 		Title:  "JUST NULLABLE",
-		Schema: openapi3.NewSchema().WithNullable(),
+		Schema: NewSchema().WithNullable(),
 		Serialization: map[string]interface{}{
 			// This OA3 schema is exactly both this draft-04 schema: {} and:
 			// {anyOf: [type:string, type:number, type:integer, type:boolean
@@ -114,7 +133,7 @@ var schemaExamples = []schemaExample{
 
 	{
 		Title:  "NULLABLE BOOLEAN",
-		Schema: openapi3.NewBoolSchema().WithNullable(),
+		Schema: NewBoolSchema().WithNullable(),
 		Serialization: map[string]interface{}{
 			"nullable": true,
 			"type":     "boolean",
@@ -136,9 +155,9 @@ var schemaExamples = []schemaExample{
 
 	{
 		Title: "NULLABLE ANYOF",
-		Schema: openapi3.NewAnyOfSchema(
-			openapi3.NewIntegerSchema(),
-			openapi3.NewFloat64Schema(),
+		Schema: NewAnyOfSchema(
+			NewIntegerSchema(),
+			NewFloat64Schema(),
 		).WithNullable(),
 		Serialization: map[string]interface{}{
 			"nullable": true,
@@ -160,9 +179,93 @@ var schemaExamples = []schemaExample{
 		},
 	},
 
+	{
+		Title: "ANYOF NULLABLE CHILD",
+		Schema: NewAnyOfSchema(
+			NewIntegerSchema().WithNullable(),
+			NewFloat64Schema(),
+		),
+		Serialization: map[string]interface{}{
+			"anyOf": []interface{}{
+				map[string]interface{}{"type": "integer", "nullable": true},
+				map[string]interface{}{"type": "number"},
+			},
+		},
+		AllValid: []interface{}{
+			nil,
+			42,
+			4.2,
+		},
+		AllInvalid: []interface{}{
+			true,
+			[]interface{}{42},
+			"bla",
+			map[string]interface{}{},
+		},
+	},
+
+	{
+		Title: "NULLABLE ALLOF",
+		Schema: NewAllOfSchema(
+			NewBoolSchema().WithNullable(),
+			NewBoolSchema().WithNullable(),
+		),
+		Serialization: map[string]interface{}{
+			"allOf": []interface{}{
+				map[string]interface{}{"type": "boolean", "nullable": true},
+				map[string]interface{}{"type": "boolean", "nullable": true},
+			},
+		},
+		AllValid: []interface{}{
+			nil,
+			true,
+			false,
+		},
+		AllInvalid: []interface{}{
+			2,
+			4.2,
+			[]interface{}{42},
+			"bla",
+			map[string]interface{}{},
+		},
+	},
+
+	{
+		Title: "ANYOF WITH PARENT CONSTRAINTS",
+		Schema: NewAnyOfSchema(
+			NewObjectSchema().WithRequired([]string{"stringProp"}),
+			NewObjectSchema().WithRequired([]string{"boolProp"}),
+		).WithProperties(map[string]*Schema{
+			"stringProp": NewStringSchema().WithMaxLength(18),
+			"boolProp":   NewBoolSchema(),
+		}),
+		Serialization: map[string]interface{}{
+			"properties": map[string]interface{}{
+				"stringProp": map[string]interface{}{"type": "string", "maxLength": 18},
+				"boolProp":   map[string]interface{}{"type": "boolean"},
+			},
+			"anyOf": []interface{}{
+				map[string]interface{}{"type": "object", "required": []string{"stringProp"}},
+				map[string]interface{}{"type": "object", "required": []string{"boolProp"}},
+			},
+		},
+		AllValid: []interface{}{
+			map[string]interface{}{"stringProp": "valid string value"},
+			map[string]interface{}{"boolProp": true},
+			map[string]interface{}{"stringProp": "valid string value", "boolProp": true},
+		},
+		AllInvalid: []interface{}{
+			1,
+			map[string]interface{}{},
+			map[string]interface{}{"invalidProp": false},
+			map[string]interface{}{"stringProp": "invalid string value"},
+			map[string]interface{}{"stringProp": "invalid string value", "boolProp": true},
+		},
+	},
+
 	{
 		Title:  "BOOLEAN",
-		Schema: openapi3.NewBoolSchema(),
+		Schema: NewBoolSchema(),
 		Serialization: map[string]interface{}{
 			"type": "boolean",
 		},
@@ -181,7 +284,7 @@ var schemaExamples = []schemaExample{
 
 	{
 		Title: "NUMBER",
-		Schema: openapi3.NewFloat64Schema().
+		Schema: NewFloat64Schema().
 			WithMin(2.5).
 			WithMax(3.5),
 		Serialization: map[string]interface{}{
@@ -208,7 +311,7 @@ var schemaExamples = []schemaExample{
 
 	{
 		Title: "INTEGER",
-		Schema: openapi3.NewInt64Schema().
+		Schema: NewInt64Schema().
 			WithMin(2).
 			WithMax(5),
 		Serialization: map[string]interface{}{
@@ -233,10 +336,59 @@ var schemaExamples = []schemaExample{
 			map[string]interface{}{},
 		},
 	},
-
+	{
+		Title:  "INTEGER OPTIONAL INT64 FORMAT",
+		Schema: NewInt64Schema(),
+		Serialization: map[string]interface{}{
+			"type":   "integer",
+			"format": "int64",
+		},
+		AllValid: []interface{}{
+			1,
+			256,
+			65536,
+			int64(math.MaxInt32) + 10,
+			int64(math.MinInt32) - 10,
+		},
+		AllInvalid: []interface{}{
+			nil,
+			false,
+			3.5,
+			true,
+			"",
+			[]interface{}{},
+			map[string]interface{}{},
+		},
+	},
+	{
+		Title:  "INTEGER OPTIONAL INT32 FORMAT",
+		Schema: NewInt32Schema(),
+		Serialization: map[string]interface{}{
+			"type":   "integer",
+			"format": "int32",
+		},
+		AllValid: []interface{}{
+			1,
+			256,
+			65536,
+			int64(math.MaxInt32),
+			int64(math.MaxInt32),
+		},
+		AllInvalid: []interface{}{
+			nil,
+			false,
+			3.5,
+			int64(math.MaxInt32) + 10,
+			int64(math.MinInt32) - 10,
+			true,
+			"",
+			[]interface{}{},
+			map[string]interface{}{},
+		},
+	},
 	{
 		Title: "STRING",
-		Schema: openapi3.NewStringSchema().
+		Schema: NewStringSchema().
 			WithMinLength(2).
 			WithMaxLength(3).
 			WithPattern("^[abc]+$"),
@@ -265,7 +417,7 @@ var schemaExamples = []schemaExample{
 
 	{
 		Title:  "STRING: optional format 'uuid'",
-		Schema: openapi3.NewUUIDSchema(),
+		Schema: NewUUIDSchema(),
 		Serialization: map[string]interface{}{
 			"type":   "string",
 			"format": "uuid",
@@ -274,6 +426,12 @@ var schemaExamples = []schemaExample{
 			"dd7d8481-81a3-407f-95f0-a2f1cb382a4b",
 			"dcba3901-2fba-48c1-9db2-00422055804e",
 			"ace8e3be-c254-4c10-8859-1401d9a9d52a",
+			"DD7D8481-81A3-407F-95F0-A2F1CB382A4B",
+			"DCBA3901-2FBA-48C1-9DB2-00422055804E",
+			"ACE8E3BE-C254-4C10-8859-1401D9A9D52A",
+			"dd7D8481-81A3-407f-95F0-A2F1CB382A4B",
+			"DCBA3901-2FBA-48C1-9db2-00422055804e",
+			"ACE8E3BE-c254-4C10-8859-1401D9A9D52A",
 		},
 		AllInvalid: []interface{}{
 			nil,
@@ -281,12 +439,19 @@ var schemaExamples = []schemaExample{
 			"4cf3i040-ea14-4daa-b0b5-ea9329473519",
 			"aaf85740-7e27-4b4f-b4554-a03a43b1f5e3",
 			"56f5bff4-z4b6-48e6-a10d-b6cf66a83b04",
+			"G39840B1-D0EF-446D-E555-48FCCA50A90A",
+			"4CF3I040-EA14-4DAA-B0B5-EA9329473519",
+			"AAF85740-7E27-4B4F-B4554-A03A43B1F5E3",
+			"56F5BFF4-Z4B6-48E6-A10D-B6CF66A83B04",
+			"4CF3I040-EA14-4Daa-B0B5-EA9329473519",
+			"AAf85740-7E27-4B4F-B4554-A03A43b1F5E3",
+			"56F5Bff4-Z4B6-48E6-a10D-B6CF66A83B04",
 		},
 	},
 
 	{
 		Title:  "STRING: format 'date-time'",
-		Schema: openapi3.NewDateTimeSchema(),
+		Schema: NewDateTimeSchema(),
 		Serialization: map[string]interface{}{
 			"type":   "string",
 			"format": "date-time",
@@ -311,7 +476,7 @@ var schemaExamples = []schemaExample{
 
 	{
 		Title:  "STRING: format 'date-time'",
-		Schema: openapi3.NewBytesSchema(),
+		Schema: NewBytesSchema(),
 		Serialization: map[string]interface{}{
 			"type":   "string",
 			"format": "byte",
@@ -336,19 +501,19 @@ var schemaExamples = []schemaExample{
 		AllInvalid: []interface{}{
 			nil,
 			" ",
-			"\n",
+			"\n\n", // a \n is ok for JSON but not for YAML decoder/encoder
 			"%",
 		},
 	},
 
 	{
 		Title: "ARRAY",
-		Schema: &openapi3.Schema{
+		Schema: &Schema{
 			Type:        "array",
 			MinItems:    2,
-			MaxItems:    openapi3.Uint64Ptr(3),
+			MaxItems:    Uint64Ptr(3),
 			UniqueItems: true,
-			Items:       openapi3.NewFloat64Schema().NewRef(),
+			Items:       NewFloat64Schema().NewRef(),
 		},
 		Serialization: map[string]interface{}{
 			"type":        "array",
@@ -383,13 +548,13 @@ var schemaExamples = []schemaExample{
 	},
 	{
 		Title: "ARRAY : items format 'object'",
-		Schema: &openapi3.Schema{
+		Schema: &Schema{
 			Type:        "array",
 			UniqueItems: true,
-			Items: (&openapi3.Schema{
+			Items: (&Schema{
 				Type: "object",
-				Properties: map[string]*openapi3.SchemaRef{
-					"key1": openapi3.NewFloat64Schema().NewRef(),
+				Properties: Schemas{
+					"key1": NewFloat64Schema().NewRef(),
 				},
 			}).NewRef(),
 		},
@@ -410,7 +575,7 @@ var schemaExamples = []schemaExample{
 				map[string]interface{}{
 					"key1": 1,
 					"key2": 1,
-					// Additioanl properties will make object different
+					// Additional properties will make object different
 					// By default additionalProperties is true
 				},
 				map[string]interface{}{
@@ -440,16 +605,16 @@ var schemaExamples = []schemaExample{
 
 	{
 		Title: "ARRAY : items format 'object' and object with a property of array type ",
-		Schema: &openapi3.Schema{
+		Schema: &Schema{
 			Type:        "array",
 			UniqueItems: true,
-			Items: (&openapi3.Schema{
+			Items: (&Schema{
 				Type: "object",
-				Properties: map[string]*openapi3.SchemaRef{
-					"key1": (&openapi3.Schema{
+				Properties: Schemas{
+					"key1": (&Schema{
 						Type:        "array",
 						UniqueItems: true,
-						Items:       openapi3.NewFloat64Schema().NewRef(),
+						Items:       NewFloat64Schema().NewRef(),
 					}).NewRef(),
 				},
 			}).NewRef(),
@@ -526,13 +691,13 @@ var schemaExamples = []schemaExample{
 
 	{
 		Title: "ARRAY : items format 'array'",
-		Schema: &openapi3.Schema{
+		Schema: &Schema{
 			Type:        "array",
 			UniqueItems: true,
-			Items: (&openapi3.Schema{
+			Items: (&Schema{
 				Type:        "array",
 				UniqueItems: true,
-				Items:       openapi3.NewFloat64Schema().NewRef(),
+				Items:       NewFloat64Schema().NewRef(),
 			}).NewRef(),
 		},
 		Serialization: map[string]interface{}{
@@ -570,16 +735,16 @@ var schemaExamples = []schemaExample{
 
 	{
 		Title: "ARRAY : items format 'array' and array with object type items",
-		Schema: &openapi3.Schema{
+		Schema: &Schema{
 			Type:        "array",
 			UniqueItems: true,
-			Items: (&openapi3.Schema{
+			Items: (&Schema{
 				Type:        "array",
 				UniqueItems: true,
-				Items: (&openapi3.Schema{
+				Items: (&Schema{
 					Type: "object",
-					Properties: map[string]*openapi3.SchemaRef{
-						"key1": openapi3.NewFloat64Schema().NewRef(),
+					Properties: Schemas{
+						"key1": NewFloat64Schema().NewRef(),
 					},
 				}).NewRef(),
 			}).NewRef(),
@@ -674,11 +839,11 @@ var schemaExamples = []schemaExample{
 
 	{
 		Title: "OBJECT",
-		Schema: &openapi3.Schema{
+		Schema: &Schema{
 			Type:     "object",
-			MaxProps: openapi3.Uint64Ptr(2),
-			Properties: map[string]*openapi3.SchemaRef{
-				"numberProperty": openapi3.NewFloat64Schema().NewRef(),
+			MaxProps: Uint64Ptr(2),
+			Properties: Schemas{
+				"numberProperty": NewFloat64Schema().NewRef(),
 			},
 		},
 		Serialization: map[string]interface{}{
@@ -718,13 +883,13 @@ var schemaExamples = []schemaExample{
 		},
 	},
 	{
-		Schema: &openapi3.Schema{
+		Schema: &Schema{
 			Type: "object",
-			AdditionalProperties: &openapi3.SchemaRef{
-				Value: &openapi3.Schema{
+			AdditionalProperties: AdditionalProperties{Schema: &SchemaRef{
+				Value: &Schema{
 					Type: "number",
 				},
-			},
+			}},
 		},
 		Serialization: map[string]interface{}{
 			"type": "object",
@@ -746,9 +911,9 @@ var schemaExamples = []schemaExample{
 		},
 	},
 	{
-		Schema: &openapi3.Schema{
-			Type:                        "object",
-			AdditionalPropertiesAllowed: openapi3.BoolPtr(true),
+		Schema: &Schema{
+			Type:                 "object",
+			AdditionalProperties: AdditionalProperties{Has: BoolPtr(true)},
 		},
 		Serialization: map[string]interface{}{
 			"type":                 "object",
@@ -765,9 +930,9 @@ var schemaExamples = []schemaExample{
 
 	{
 		Title: "NOT",
-		Schema: &openapi3.Schema{
-			Not: &openapi3.SchemaRef{
-				Value: &openapi3.Schema{
+		Schema: &Schema{
+			Not: &SchemaRef{
+				Value: &Schema{
 					Enum: []interface{}{
 						nil,
 						true,
@@ -802,15 +967,15 @@ var schemaExamples = []schemaExample{
 
 	{
 		Title: "ANY OF",
-		Schema: &openapi3.Schema{
-			AnyOf: []*openapi3.SchemaRef{
+		Schema: &Schema{
+			AnyOf: []*SchemaRef{
 				{
-					Value: openapi3.NewFloat64Schema().
+					Value: NewFloat64Schema().
 						WithMin(1).
 						WithMax(2),
 				},
 				{
-					Value: openapi3.NewFloat64Schema().
+					Value: NewFloat64Schema().
 						WithMin(2).
 						WithMax(3),
 				},
@@ -843,15 +1008,15 @@ var schemaExamples = []schemaExample{
 
 	{
 		Title: "ALL OF",
-		Schema: &openapi3.Schema{
-			AllOf: []*openapi3.SchemaRef{
+		Schema: &Schema{
+			AllOf: []*SchemaRef{
 				{
-					Value: openapi3.NewFloat64Schema().
+					Value: NewFloat64Schema().
 						WithMin(1).
 						WithMax(2),
 				},
 				{
-					Value: openapi3.NewFloat64Schema().
+					Value: NewFloat64Schema().
 						WithMin(2).
 						WithMax(3),
 				},
@@ -884,15 +1049,15 @@ var schemaExamples = []schemaExample{
 
 	{
 		Title: "ONE OF",
-		Schema: &openapi3.Schema{
-			OneOf: []*openapi3.SchemaRef{
+		Schema: &Schema{
+			OneOf: []*SchemaRef{
 				{
-					Value: openapi3.NewFloat64Schema().
+					Value: NewFloat64Schema().
 						WithMin(1).
 						WithMax(2),
 				},
 				{
-					Value: openapi3.NewFloat64Schema().
+					Value: NewFloat64Schema().
 						WithMin(2).
 						WithMax(3),
 				},
@@ -926,7 +1091,7 @@ var schemaExamples = []schemaExample{
 
 type schemaTypeExample struct {
 	Title      string
-	Schema     *openapi3.Schema
+	Schema     *Schema
 	AllValid   []string
 	AllInvalid []string
 }
@@ -942,12 +1107,13 @@ func testType(t *testing.T, example schemaTypeExample) func(*testing.T) {
 		baseSchema := example.Schema
 		for _, typ := range example.AllValid {
 			schema := baseSchema.WithFormat(typ)
-			err := schema.Validate(context.TODO())
+			err := schema.Validate(context.Background())
 			require.NoError(t, err)
 		}
 		for _, typ := range example.AllInvalid {
 			schema := baseSchema.WithFormat(typ)
-			err := schema.Validate(context.TODO())
+			ctx := WithValidationOptions(context.Background(), EnableSchemaFormatValidation())
+			err := schema.Validate(ctx)
 			require.Error(t, err)
 		}
 	}
@@ -956,7 +1122,7 @@ func testType(t *testing.T, example schemaTypeExample) func(*testing.T) {
 var typeExamples = []schemaTypeExample{
 	{
 		Title:  "STRING",
-		Schema: openapi3.NewStringSchema(),
+		Schema: NewStringSchema(),
 		AllValid: []string{
 			"",
 			"byte",
@@ -974,7 +1140,7 @@ var typeExamples = []schemaTypeExample{
 
 	{
 		Title:  "NUMBER",
-		Schema: openapi3.NewFloat64Schema(),
+		Schema: NewFloat64Schema(),
 		AllValid: []string{
 			"",
 			"float",
@@ -987,7 +1153,7 @@ var typeExamples = []schemaTypeExample{
 
 	{
 		Title:  "INTEGER",
-		Schema: openapi3.NewIntegerSchema(),
+		Schema: NewIntegerSchema(),
 		AllValid: []string{
 			"",
 			"int32",
@@ -1014,29 +1180,29 @@ func testSchemaError(t *testing.T, example schemaErrorExample) func(*testing.T)
 
 type schemaErrorExample struct {
 	Title string
-	Error *openapi3.SchemaError
+	Error *SchemaError
 	Want  string
 }
 
 var schemaErrorExamples = []schemaErrorExample{
 	{
 		Title: "SIMPLE",
-		Error: &openapi3.SchemaError{
+		Error: &SchemaError{
 			Value:  1,
-			Schema: &openapi3.Schema{},
+			Schema: &Schema{},
 			Reason: "SIMPLE",
 		},
 		Want: "SIMPLE",
 	},
 	{
 		Title: "NEST",
-		Error: &openapi3.SchemaError{
+		Error: &SchemaError{
 			Value:  1,
-			Schema: &openapi3.Schema{},
+			Schema: &Schema{},
 			Reason: "PARENT",
-			Origin: &openapi3.SchemaError{
+			Origin: &SchemaError{
 				Value:  1,
-				Schema: &openapi3.Schema{},
+				Schema: &Schema{},
 				Reason: "NEST",
 			},
 		},
@@ -1044,30 +1210,252 @@ var schemaErrorExamples = []schemaErrorExample{
 	},
 }
 
-func TestRegisterArrayUniqueItemsChecker(t *testing.T) {
-	var (
-		checker = func(items []interface{}) bool {
-			return false
-		}
-		scheme = openapi3.Schema{
-			Type:        "array",
-			UniqueItems: true,
-			Items:       openapi3.NewStringSchema().NewRef(),
-		}
-		val = []interface{}{"1", "2", "3"}
-		err error
-	)
+type schemaMultiErrorExample struct {
+	Title          string
+	Schema         *Schema
+	Values         []interface{}
+	ExpectedErrors []MultiError
+}
 
-	// Fist checked by predefined function
-	err = scheme.VisitJSON(val)
+func TestSchemasMultiError(t *testing.T) {
+	for _, example := range schemaMultiErrorExamples {
+		t.Run(example.Title, testSchemaMultiError(t, example))
+	}
+}
+
+func testSchemaMultiError(t *testing.T, example schemaMultiErrorExample) func(*testing.T) {
+	return func(t *testing.T) {
+		schema := example.Schema
+		for validateFuncIndex, validateFunc := range validateSchemaFuncs {
+			for i, value := range example.Values {
+				err := validateFunc(t, schema, value, MultiErrors())
+				require.Errorf(t, err, "ValidateFunc #%d, value #%d: %#", validateFuncIndex, i, value)
+				require.IsType(t, MultiError{}, err)
+
+				merr, _ := err.(MultiError)
+				expected := example.ExpectedErrors[i]
+				require.True(t, len(merr) > 0)
+				require.Len(t, merr, len(expected))
+				for _, e := range merr {
+					require.IsType(t, &SchemaError{}, e)
+					var found bool
+					scherr, _ := e.(*SchemaError)
+					for _, expectedErr := range expected {
+						expectedScherr, _ := expectedErr.(*SchemaError)
+						if reflect.DeepEqual(expectedScherr.reversePath, scherr.reversePath) &&
+							expectedScherr.SchemaField == scherr.SchemaField {
+							found = true
+							break
+						}
+					}
+					require.Truef(t, found, "ValidateFunc #%d, value #%d: missing %s error on %s", validateFunc, i, scherr.SchemaField, strings.Join(scherr.JSONPointer(), "."))
+				}
+			}
+		}
+	}
+}
+
+var schemaMultiErrorExamples = []schemaMultiErrorExample{
+	{
+		Title: "STRING",
+		Schema: NewStringSchema().
+			WithMinLength(2).
+			WithMaxLength(3).
+			WithPattern("^[abc]+$"),
+		Values: []interface{}{
+			"f",
+			"foobar",
+		},
+		ExpectedErrors: []MultiError{
+			{&SchemaError{SchemaField: "minLength"}, &SchemaError{SchemaField: "pattern"}},
+			{&SchemaError{SchemaField: "maxLength"}, &SchemaError{SchemaField: "pattern"}},
+		},
+	},
+	{
+		Title: "NUMBER",
+		Schema: NewIntegerSchema().
+			WithMin(1).
+			WithMax(10),
+		Values: []interface{}{
+			0.5,
+			10.1,
+		},
+		ExpectedErrors: []MultiError{
+			{&SchemaError{SchemaField: "type"}, &SchemaError{SchemaField: "minimum"}},
+			{&SchemaError{SchemaField: "type"}, &SchemaError{SchemaField: "maximum"}},
+		},
+	},
+	{
+		Title: "ARRAY: simple",
+		Schema: NewArraySchema().
+			WithMinItems(2).
+			WithMaxItems(2).
+			WithItems(NewStringSchema().
+				WithPattern("^[abc]+$")),
+		Values: []interface{}{
+			[]interface{}{"foo"},
+			[]interface{}{"foo", "bar", "fizz"},
+		},
+		ExpectedErrors: []MultiError{
+			{
+				&SchemaError{SchemaField: "minItems"},
+				&SchemaError{SchemaField: "pattern", reversePath: []string{"0"}},
+			},
+			{
+				&SchemaError{SchemaField: "maxItems"},
+				&SchemaError{SchemaField: "pattern", reversePath: []string{"0"}},
+				&SchemaError{SchemaField: "pattern", reversePath: []string{"1"}},
+				&SchemaError{SchemaField: "pattern", reversePath: []string{"2"}},
+			},
+		},
+	},
+	{
+		Title: "ARRAY: object",
+		Schema: NewArraySchema().
+			WithItems(NewObjectSchema().
+				WithProperties(map[string]*Schema{
+					"key1": NewStringSchema(),
+					"key2": NewIntegerSchema(),
+				}),
+			),
+		Values: []interface{}{
+			[]interface{}{
+				map[string]interface{}{
+					"key1": 100, // not a string
+					"key2": "not an integer",
+				},
+			},
+		},
+		ExpectedErrors: []MultiError{
+			{
+				&SchemaError{SchemaField: "type", reversePath: []string{"key1", "0"}},
+				&SchemaError{SchemaField: "type", reversePath: []string{"key2", "0"}},
+			},
+		},
+	},
+	{
+		Title: "OBJECT",
+		Schema: NewObjectSchema().
+			WithProperties(map[string]*Schema{
+				"key1": NewStringSchema(),
+				"key2": NewIntegerSchema(),
+				"key3": NewArraySchema().
+					WithItems(NewStringSchema().
+						WithPattern("^[abc]+$")),
+			}),
+		Values: []interface{}{
+			map[string]interface{}{
+				"key1": 100, // not a string
+				"key2": "not an integer",
+				"key3": []interface{}{"abc", "def"},
+			},
+		},
+		ExpectedErrors: []MultiError{
+			{
+				&SchemaError{SchemaField: "type", reversePath: []string{"key1"}},
+				&SchemaError{SchemaField: "type", reversePath: []string{"key2"}},
+				&SchemaError{SchemaField: "pattern", reversePath: []string{"1", "key3"}},
+			},
+		},
+	},
+}
+
+func TestIssue283(t *testing.T) {
+	spec := []byte(`
+openapi: "3.0.1"
+paths: {}
+info:
+  version: 1.1.1
+  title: title
+components:
+  schemas:
+    Test:
+      properties:
+        name:
+          type: string
+        ownerName:
+          not:
+            type: boolean
+      type: object
+`[1:])
+
+	data := map[string]interface{}{
+		"name":      "kin-openapi",
+		"ownerName": true,
+	}
+
+	loader := NewLoader()
+	doc, err := loader.LoadFromData(spec)
+	require.NoError(t, err)
+	err = doc.Validate(loader.Context)
 	require.NoError(t, err)
 
-	// Register a function will always return false when check if a
-	// slice has unique items, then use a slice indeed has unique
-	// items to verify that check unique items will failed.
-	openapi3.RegisterArrayUniqueItemsChecker(checker)
+	err = doc.Components.Schemas["Test"].Value.VisitJSON(data)
+	require.NotNil(t, err)
+	require.NotEqual(t, errSchema, err)
+	require.ErrorContains(t, err, `Error at "/ownerName": Doesn't match schema "not"`)
+}
 
-	err = scheme.VisitJSON(val)
+func TestValidationFailsOnInvalidPattern(t *testing.T) {
+	schema := Schema{
+		Pattern: "[",
+		Type:    "string",
+	}
+
+	err := schema.Validate(context.Background())
+	require.Error(t, err)
+}
+
+func TestIssue646(t *testing.T) {
+	data := []byte(`
+enum:
+- 42
+- []
+- [a]
+- {}
+- {b: c}
+`[1:])
+
+	var schema Schema
+	err := yaml.Unmarshal(data, &schema)
+	require.NoError(t, err)
+
+	err = schema.Validate(context.Background())
+	require.NoError(t, err)
+
+	err = schema.VisitJSON(42)
+	require.NoError(t, err)
+
+	err = schema.VisitJSON(1337)
+	require.Error(t, err)
+
+	err = schema.VisitJSON([]interface{}{})
+	require.NoError(t, err)
+
+	err = schema.VisitJSON([]interface{}{"a"})
+	require.NoError(t, err)
+
+	err = schema.VisitJSON([]interface{}{"b"})
+	require.Error(t, err)
+
+	err = schema.VisitJSON(map[string]interface{}{})
+	require.NoError(t, err)
+
+	err = schema.VisitJSON(map[string]interface{}{"b": "c"})
+	require.NoError(t, err)
+
+	err = schema.VisitJSON(map[string]interface{}{"d": "e"})
 	require.Error(t, err)
-	require.True(t, strings.HasPrefix(err.Error(), "Duplicate items found"))
+}
+
+func TestIssue751(t *testing.T) {
+	schema := &Schema{
+		Type:        "array",
+		UniqueItems: true,
+		Items:       NewStringSchema().NewRef(),
+	}
+	validData := []string{"foo", "bar"}
+	invalidData := []string{"foo", "foo"}
+	require.NoError(t, schema.VisitJSON(validData))
+	require.ErrorContains(t, schema.VisitJSON(invalidData), "duplicate items found")
 }
